gt_description = "Source Logic Tree v9.0.0"


logic_tree_permutations = [
    [
        {
            "tag": "Final SLT for TAG Workshop - Hik fix, Cru fix, Puy fix",
            "weight": 1.0,
            "permute": [
                {
                    "group": "CRU",
                    "nrlz": 21,
                    "members": [
                        {
                            "tag": "geologic, TI, N4.6, b1.089 C4.2 s1",
                            "weight": 1.0,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE0NzU=",
                            "bg_id": "RmlsZToxMzA3Mjg=",
                        },
                    ],
                },
            ],
        }
    ]
]



gt_description = "Source Logic Tree v9.0.0"

logic_tree_permutations = [
    [
        {
            "tag": "Final SLT for TAG Workshop - Hik fix, Cru fix, Puy fix",
            "weight": 1.0,
            "permute": [
                {
                    "group": "Hik",
                    "nrlz": 12,
                    "members": [
                        {
                            "tag": "Hik TL, N16.5, b0.95, C4, s0.42",
                            "weight": 1.0,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE2MTE=",
                            "bg_id": "RmlsZToxMzA3MzI=",
                        },
                    ],
                },
                {
                    "group": "PUY",
                    "nrlz": 12,
                    "members": [
                        {
                            "tag": "Puy 0.7, N4.6, b0.902, C4, s0.28",
                            "weight": 1.0,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE2NTc=",
                            "bg_id": "RmlsZToxMzA3NTM=",
                        },
                    ],
                },
                {
                    "group": "CRU",
                    "nrlz": 21,
                    "members": [
                        {
                            "tag": "geologic, TI, N3.4, b0.959 C4.2 s0.66",
                            "weight": 1.0,
                            "inv_id": "SW52ZXJzaW9uU29sdXRpb25Ocm1sOjEyOTE0NjQ=",
                            "bg_id": "RmlsZToxMzA3MTY=",
                        },
                    ],
                },
                {
                    "group": "SLAB",
                    "nrlz": 12,
                    "members": [
                        {"tag": "slab-uniform-1depth-rates", "weight": 1.0, "inv_id": "", "bg_id": "RmlsZToxMjEwMzM="}
                    ],
                },
            ],
        }
    ]
]



"""
Classes for deserialising NRML XML hazard uncertainty models.
"""
from dataclasses import dataclass, field
from pathlib import PurePath
from typing import TYPE_CHECKING, Iterator, List

from lxml import objectify

if TYPE_CHECKING:
    from nzshm_model.source_logic_tree.version2 import logic_tree as slt

    from .logic_tree import LogicTreeBranch


def _strip_whitespace(string: str) -> str:
    return ''.join(string.split())


@dataclass
class GenericUncertaintyModel:
    parent: "LogicTreeBranch"
    text: str = ""

    @classmethod
    def from_parent_element(cls, node: objectify.Element, parent: "LogicTreeBranch") -> "GenericUncertaintyModel":
        return GenericUncertaintyModel(parent=parent, text=node.text.strip())

    def path(self) -> PurePath:
        return PurePath(self.parent.path(), _strip_whitespace(self.text))


@dataclass
class GroundMotionUncertaintyModel:
    parent: "LogicTreeBranch"
    gmpe_name: str
    arguments: List[str]
    text: str = ""

    @classmethod
    def from_parent_element(cls, node: objectify.Element, parent: "LogicTreeBranch") -> "GroundMotionUncertaintyModel":
        lines = node.text.split("\n")
        return GroundMotionUncertaintyModel(
            parent=parent,
            text=node.text.strip(),
            gmpe_name=lines[0].strip(),
            arguments=[arg.strip() for arg in lines[1:]],
        )

    def path(self) -> PurePath:
        return PurePath(self.parent.path(), _strip_whitespace(self.text))  # todo unique combination of name + args


@dataclass
class SourcesUncertaintyModel:
    parent: "LogicTreeBranch"
    source_files: List[str]
    text: str = ""

    @classmethod
    def from_parent_element(cls, node: objectify.Element, parent: "LogicTreeBranch") -> "SourcesUncertaintyModel":
        lines = node.text.split()  # splitting on whitespace
        return SourcesUncertaintyModel(
            parent=parent, text=node.text.strip(), source_files=[arg.strip() for arg in lines]
        )

    def path(self) -> PurePath:
        return PurePath(self.parent.path(), _strip_whitespace(self.text))  # todo unique combination of sources


@dataclass
class NshmSourceUncertaintyModel:
    parent: "LogicTreeBranch"
    source_files: List[str] = field(
        default_factory=list
    )  # deferred , we must unpack these from the nrml_id (either direct source or zip file containing sources)
    toshi_nrml_id: str = ""  # in NZSHM22 this is toshi ID
    global_uri: str = ""  # eventually NZSHM or others may publish these as URI resources.
    model_type: str = ""

    @classmethod
    def from_parent_slt(
        cls, ltb: "slt.SourceBranch", parent: "LogicTreeBranch"
    ) -> Iterator["NshmSourceUncertaintyModel"]:
        """resolve to filenames of NRML sources"""
        for source in ltb.sources:
            # print(source)
            if source.type == 'inversion':
                yield NshmSourceUncertaintyModel(parent=parent, toshi_nrml_id=source.nrml_id, model_type="FaultSource")
            if source.type == 'distributed':
                yield NshmSourceUncertaintyModel(
                    parent=parent, toshi_nrml_id=source.nrml_id, model_type="DistributedSource"
                )

    def path(self) -> PurePath:
        return PurePath(self.parent.path(), _strip_whitespace(self.toshi_nrml_id))  # todo unique combination of sources



"""
The model version module defines the availabe models

Attributes:
    versions (Dict[str, str]): a mapping from model version string to its module name

"""
from typing import List

from .model import NshmModel, versions

CURRENT_VERSION = "NSHM_v1.0.4"


def all_model_versions() -> List[str]:
    """
    get the list of available model versions
    """
    return list(versions.keys())


def get_model_version(version: str = CURRENT_VERSION) -> 'NshmModel':
    """
    A simple wrapper for the underlying NshmModel static method

    Returns:
        the model instance.
    """
    return NshmModel.get_model_version(version)



"""
Classes for defining logic tree branches
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import reduce
from operator import mul
from typing import TYPE_CHECKING, Sequence

if TYPE_CHECKING:
    from .logic_tree_base import BranchSet, FilteredBranch, LogicTree


@dataclass
class Branch(ABC):
    """
    Abstract baseclass for logic tree branches

    Arguments:
        name: a name for the branch
        weight: a weight for the branch
    """

    branch_id: str = ""
    weight: float = 1.0

    @abstractmethod
    def filtered_branch(self, logic_tree: 'LogicTree', branch_set: 'BranchSet') -> 'FilteredBranch':
        """
        Produce a new filtered branch with the properties of the branch

        Parameters:
            logic_tree: The logic tree that the branch belongs to
            branch_set: The branch est that the branch belongs to

        Returns:
            a filtered branch
        """
        pass


@dataclass
class CompositeBranch:
    """
    A logic tree branch comprised of combinations of branches from one or more branch sets.

    Arguments:
        branches: the component branches (branches from branch sets) in the composite branch
        weight: the weight of the composite branch
    """

    branches: Sequence[Branch] = field(default_factory=list)
    weight: float = 1.0

    def __post_init__(self) -> None:
        self.weight = reduce(mul, [branch.weight for branch in self.branches], 1.0)

    def __iter__(self):
        self.__counter = 0
        return self

    def __next__(self):
        if self.__counter >= len(self.branches):
            raise StopIteration
        else:
            self.__counter += 1
            return self.branches[self.__counter - 1]



# import pytest

# from nzshm_model.branch_prop import get_source_branch_sets, get_source_branches


# def test_get_source_branch_sets_with_list():
#     branch_sets = list(get_source_branch_sets('NSHM_v1.0.4', ['CRU', 'PUY']))
#     assert len(branch_sets) == 2
#     assert branch_sets[0].short_name == 'PUY'
#     assert branch_sets[1].short_name == 'CRU'


# def test_get_source_branch_sets_with_single():
#     branch_sets = list(get_source_branch_sets('NSHM_v1.0.4', 'CRU'))
#     assert len(branch_sets) == 1
#     assert branch_sets[0].short_name == 'CRU'


# def test_get_source_branch_sets_with_null():
#     assert len(list(get_source_branch_sets('NSHM_v1.0.4', []))) == 4
#     assert len(list(get_source_branch_sets('NSHM_v1.0.4'))) == 4


# def test_get_source_branch_sets_invalid_model_version():
#     with pytest.raises(ValueError, match=r'.* is not a valid'):
#         next(get_source_branch_sets('', ['CRU', 'PUY']))


# def test_get_source_branch_sets_with_invalid_branch():
#     with pytest.raises(StopIteration):
#         next(get_source_branch_sets('NSHM_v1.0.4', ['XXX']))


# def test_get_first_crustal_branches():
#     cru_branches = next(get_source_branch_sets('NSHM_v1.0.4', 'CRU')).branches
#     assert next(get_source_branches('NSHM_v1.0.4', 'CRU')) == cru_branches[0]


# def test_get_source_branches_invalid_model_version():
#     with pytest.raises(ValueError, match=r'.* is not a valid'):
#         # a
#         next(get_source_branches('', ['CRU', 'PUY']))

#     with pytest.raises(ValueError, match=r'.* is not a valid'):
#         # b
#         list(get_source_branches('', ['CRU', 'PUY']))

#     with pytest.raises(ValueError, match=r'.* is not a valid'):
#         # c for next
#         for b in get_source_branches('', ['CRU', 'PUY']):
#             pass


# def test_get_source_branches_with_invalid_branch():
#     with pytest.raises(StopIteration):
#         next(get_source_branches('NSHM_v1.0.4', ['XXX']))



import json
from pathlib import Path

from nzshm_model.logic_tree import SourceLogicTree
from nzshm_model.logic_tree.source_logic_tree import SourceLogicTreeV1
from nzshm_model.logic_tree.source_logic_tree.version1.logic_tree import FlattenedSourceLogicTree


def test_v2_correlations():

    slt_filepath = Path(__file__).parent.parent / "resources" / "SRM_JSON" / "nshm_v1.0.4.json"
    with slt_filepath.open() as slt_file:
        slt_v1 = SourceLogicTreeV1.from_dict(json.load(slt_file))
    slt_v2 = SourceLogicTree.from_source_logic_tree(slt_v1)

    assert len(FlattenedSourceLogicTree.from_source_logic_tree(slt_v1).branches) == len(list(slt_v2.combined_branches))
    for branch_v1, branch_v2 in zip(
        FlattenedSourceLogicTree.from_source_logic_tree(slt_v1).branches, slt_v2.combined_branches
    ):
        values_v1 = [b.values for b in branch_v1.branches]
        values_v2 = [b.values for b in branch_v2.branches]
        assert values_v1 == values_v2
        assert branch_v1.weight == branch_v2.weight



"""
test migration from slt v1 to v2
"""
from pathlib import Path

import pytest

import nzshm_model
from nzshm_model.logic_tree import SourceBranch, SourceBranchSet
from nzshm_model.logic_tree.source_logic_tree import SourceLogicTree, SourceLogicTreeV1


@pytest.fixture(scope='module')
def slt_version_1():
    # get a published v1 logic tree
    data = nzshm_model.get_model_version('NSHM_v1.0.4')._data
    yield SourceLogicTreeV1.from_dict(data)


@pytest.fixture(scope='module')
def slt_version_2():
    config = Path(__file__).parent / 'fixtures' / 'source_logic_tree_sample_0.json'
    yield SourceLogicTree.from_json(config)


@pytest.fixture(scope='module')
def full_slt():
    yield nzshm_model.get_model_version('NSHM_v1.0.4').source_logic_tree()


def test_source_logic_tree_v2_is_iterable(slt_version_2):
    for obj in slt_version_2:
        assert obj


def test_source_logic_tree_v2_is_slt_iterable(slt_version_2):
    for obj in slt_version_2:
        assert isinstance(obj, SourceBranch)
        assert isinstance(obj.branch_set, SourceBranchSet)
        assert isinstance(obj.logic_tree, SourceLogicTree)


def test_source_logic_tree_v2_iterable_size(slt_version_2):
    assert len(list(slt_version_2)) == 1
    assert len(next(iter(slt_version_2)).sources) == 2


def test_source_logic_tree_v2_iterates_source_logic_tree(slt_version_2):
    assert next(iter(slt_version_2)).sources[0] == slt_version_2.branch_sets[0].branches[0].sources[0]



#! test_module_import
import pytest

import nzshm_model


def test_current_version():
    assert nzshm_model.CURRENT_VERSION == "NSHM_v1.0.4"


def test_available_versions():
    assert len(nzshm_model.versions) == 2


@pytest.mark.parametrize("model, model_version", [('NSHM_v1.0.0', 'NSHM_v1.0.0'), ('NSHM_v1.0.4', 'NSHM_v1.0.4')])
def test_version_config(model, model_version):
    mod = nzshm_model.NshmModel.get_model_version(model)
    # mod = nzshm_model.get_model_version(model)

    assert mod.version == model_version

    assert mod.slt_config is not None
    assert mod.slt_config.logic_tree_permutations is not None


def test_get_model_version_unknown():
    with pytest.raises(ValueError, match='XXX is not a valid model version.'):
        nzshm_model.get_model_version('XXX')


@pytest.mark.parametrize('model_version', [ver for ver in nzshm_model.versions.keys()])
def test_source_logic_tree_in_all_models(model_version):
    current_model = nzshm_model.get_model_version(model_version)
    slt = current_model.source_logic_tree

    old_api_fslt = slt.fault_system_lts[0]
    new_api_fslt = slt.branch_sets[0]

    assert old_api_fslt is new_api_fslt



"""Console script for model users...."""
# noqa
import logging
import os

import click

import nzshm_model

# from nzshm_model.source_logic_tree import SourceLogicTree
from nzshm_model.logic_tree.source_logic_tree import SourceLogicTree

# from nzshm_model.source_logic_tree.slt_config import from_config, resolve_toshi_source_ids  # noqa
from nzshm_model.psha_adapter.openquake import OpenquakeSimplePshaAdapter

log = logging.getLogger()
logging.basicConfig(level=logging.DEBUG)
logging.getLogger('nshm_toshi_client.toshi_client_base').setLevel(logging.INFO)
logging.getLogger('urllib3').setLevel(logging.INFO)
logging.getLogger('botocore').setLevel(logging.INFO)
logging.getLogger('gql.transport.requests').setLevel(logging.WARN)

formatter = logging.Formatter(fmt='%(asctime)s %(levelname)-8s %(name)s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
root_handler = log.handlers[0]
root_handler.setFormatter(formatter)

log.debug('DEBUG message')
log.info('INFO message')


#  _ __ ___   __ _(_)_ __
# | '_ ` _ \ / _` | | '_ \
# | | | | | | (_| | | | | |
# |_| |_| |_|\__,_|_|_| |_|
@click.group()
def cli():
    """Nzshm-model helpers for model consumers."""


@cli.command()
@click.option('--cache_folder', '-w', default=lambda: os.getcwd())
@click.option('--model_id', '-m', default="NSHM_v1.0.4")
# @click.option('--long_filenames', '-lf', is_flag=True, help="use long filenames, instead of folders")
def fetch(cache_folder, model_id):
    """Fetch SLT sources from toshi"""
    click.echo(f"work folder: {cache_folder}")
    click.echo(f"model_id: {model_id}")

    model = nzshm_model.get_model_version(model_id)
    adapter = model.source_logic_tree().psha_adapter(provider=OpenquakeSimplePshaAdapter)

    for item in adapter.fetch_resources(cache_folder):
        click.echo(item)

    click.echo('DONE')


@cli.command()
@click.option('--cache_folder', '-w', default=lambda: os.getcwd())
@click.option('--output_folder', '-o', default=lambda: os.getcwd())
@click.option('--model_id', '-m', default="NSHM_v1.0.4")
def unpack(cache_folder, output_folder, model_id):

    model = nzshm_model.get_model_version(model_id)
    adapter = model.source_logic_tree().psha_adapter(provider=OpenquakeSimplePshaAdapter)
    source_map = adapter.unpack_resources(cache_folder, output_folder)
    click.echo(len(source_map.items()))
    click.echo('DONE')


@cli.command()
@click.option('--cache_folder', '-w', default=lambda: os.getcwd())
@click.option('--output_folder', '-o', default=lambda: os.getcwd())
@click.option('--model_id', '-m', default="NSHM_v1.0.4")
def config(cache_folder, output_folder, model_id):
    """Write an openquake hazard configuration

    This is a work in progress, as it only handles the Source LT, and not GMM LT
    or general configuration.
    """

    model = nzshm_model.get_model_version(model_id)
    slt = model.source_logic_tree  # always version 2

    ## example filter functions
    def unscaled_filter(obj):
        for v in obj.values:
            if v.name == 's' and v.value == 1.0:  # moment rate scaling
                return True

    def geodetic_filter(obj):
        for v in obj.values:
            if v.long_name == 'deformation model' and v.value == "geodetic":
                return True

    slt = SourceLogicTree.from_branches((fb for fb in slt if unscaled_filter(fb) and geodetic_filter(fb)))

    adapter = slt.psha_adapter(provider=OpenquakeSimplePshaAdapter)

    adapter.write_config(cache_folder, output_folder)
    click.echo('DONE')


if __name__ == "__main__":
    cli()  # pragma: no cover



#! nshm_v1_0_4.py

# Is this only used for the correlations ??
import nzshm_model.logic_tree.source_logic_tree.SLT_v9p0p0 as slt_config

from .model import NshmModel

model = NshmModel(  # NOQA F401
    version='NSHM_v1.0.4',
    title="NSHM version 1.0.4, corrected fault geometry",
    slt_json="nshm_v1.0.4.json",
    gmm_xml="NZ_NSHM_GMM_LT_final_EE_new_names.xml",
    slt_config=slt_config,
)



from dataclasses import dataclass, field
from typing import Any, Dict, List, Type

from nzshm_model.logic_tree.logic_tree_base import Branch, BranchSet, FilteredBranch, LogicTree
from nzshm_model.psha_adapter import PshaAdapterInterface


@dataclass
class GMCMBranch(Branch):
    gsim_name: str = ""
    gsim_args: Dict[str, Any] = field(default_factory=dict)

    def filtered_branch(self, logic_tree, branch_set):
        return GMCMFilteredBranch(logic_tree=logic_tree, branch_set=branch_set, **self.__dict__)


@dataclass
class GMCMBranchSet(BranchSet):
    tectonic_region_type: str = ""  # need a default becasue base class has a memeber with a default
    branches: List[GMCMBranch] = field(default_factory=list)


@dataclass
class GMCMLogicTree(LogicTree):
    # should we enforce that there is only one branch_set per TRT?
    branch_sets: List[GMCMBranchSet] = field(default_factory=list)

    def __post_init__(self):
        self._fix_args()

    def _fix_args(self) -> 'GMCMLogicTree':
        """Replace string representations of numeric arguments with floats"""

        def is_number(value):
            return value.replace("-", "").replace(".", "").replace("e", "").isnumeric()

        for branch_set in self.branch_sets:
            for branch in branch_set.branches:
                for k, v in branch.gsim_args.items():
                    if (isinstance(v, str)) and (is_number(v)):
                        branch.gsim_args[k] = float(v)

        return self

    def psha_adapter(self, provider: Type[PshaAdapterInterface], **kwargs):
        return provider(gmcm_logic_tree=self)

    # @classmethod
    # def from_user_config(cls, config_path: Union[Path, str]) -> 'GMCMLogicTree':
    #     raise NotImplementedError("from_user_config not implimented for GMCMLogicTree")


@dataclass
class GMCMFilteredBranch(FilteredBranch, GMCMBranch):
    logic_tree: 'GMCMLogicTree' = field(default_factory=GMCMLogicTree)
    branch_set: 'GMCMBranchSet' = field(default_factory=GMCMBranchSet)



"""
This is the top-level package for nzshm_model.

NshmModel is a general representation of an NSHM Logic Tree which includes both **Source** and
**Ground Motion Model (GMM)** logic trees.

## PSHA abstraction and adapters

PSHA applications (such as Openquake, and USGS-PHaz provide methods for calculating hazard based on input models.

PSHA adapter classes convert to/from the internal NSHM source models to the form required by a PSHA tool.
This lets us support different PSHA engines, although currently only an **openquake** adapter
is implemented.

Note that NSHM GMM logic trees use the same serialised form as openquake, so no abstraction is provided.

Todo:
 - about flattening
 - about correlations

Examples:
    >>> import nzshm_model as nm
    >>> nm.all_model_versions()
    ['NSHM_v1.0.0', 'NSHM_v1.0.4']

    >>> model = nm.get_model_version("NSHM_v1.0.4")
    >>> for branch_set in model.get_source_branch_sets('CRU'):
    >>>     for branch in branch_set.branches:
    >>>          print(branch_set.long_name, branch.weight, branch.tag)
    Crustal 0.0168335471189857 [dmgeodetic, tdFalse, bN[0.823, 2.7], C4.2, s0.66]
    ...

Attributes:
    CURRENT_VERSION (str): the version string for the currently published NSHM model.

Functions:
    all_model_versions: lists the available model versions strings
    get_model_version: Get the model instance specified

"""
from .model import NshmModel
from .model_version import CURRENT_VERSION, all_model_versions, get_model_version, versions

# Python package version is different than the NSHM MODEL version !!
__version__ = '0.9.2'



#! python test_logic_tree.py

import dataclasses
import itertools
from pathlib import Path

import pytest

from nzshm_model.logic_tree.source_logic_tree.branch_attribute import BranchAttributeSpec
from nzshm_model.logic_tree.source_logic_tree.version1.logic_tree import (
    Branch,
    BranchAttributeValue,
    CompositeBranch,
    FaultSystemLogicTree,
    FlattenedSourceLogicTree,
    SourceLogicTree,
    SourceLogicTreeCorrelation,
)
from nzshm_model.logic_tree.source_logic_tree.version1.slt_config import from_config


def test_composite_branch():
    ba0 = BranchAttributeSpec(name='C', long_name='area-magnitude scaling', value_options=[4.0])
    ba1 = BranchAttributeSpec(name='b', long_name='GR b value', value_options=[1.0])
    print(ba0, ba1)

    bav0 = BranchAttributeValue.from_branch_attribute(ba0, 4)
    bav1 = BranchAttributeValue.from_branch_attribute(ba1, 1.0)
    print(bav0, bav1)

    b0 = Branch(values=[bav0], weight=0.25, onfault_nrml_id='ABC')
    b1 = Branch(values=[bav1], weight=0.5, onfault_nrml_id='XYZ')
    print(b0, b1)

    cb = CompositeBranch([b0, b1], 10.0)
    assert cb.weight == pytest.approx(0.25 * 0.5)


def test_direct_bav():
    bav = BranchAttributeValue(name='dm', long_name='deformation model', value="geodetic")
    print(bav)
    assert bav.value == "geodetic"


def test_direct_bav_no_options():
    bav = BranchAttributeValue(name='dm', long_name='deformation model', value="geodetic")
    print(bav)
    assert bav.value == "geodetic"


def test_init():
    ba = BranchAttributeSpec(name='C', long_name='area-magnitude scaling', value_options=[4.0])
    print(ba)

    bao = BranchAttributeValue.from_branch_attribute(ba, 4)
    print(bao)

    b = Branch(values=[bao], weight=0.5, onfault_nrml_id='ABC')
    print(b)

    f = FaultSystemLogicTree('Hik', 'Hikurangi', [b])
    print(f)
    assert f.branches[0].values[0].value == 4


def build_crustal_branches():
    # define the branch attributes for our FaultSystemLogicTree
    C = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='C', long_name='area-magnitude scaling', value_options=[4])
    )
    s = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='s', long_name='moment rate scaling', value_options=[0.5, 1.0])
    )
    bN = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='bN', long_name='bN pair', value_options=[(0.87, 25), (0.95, 19.2)])
    )
    dm = BranchAttributeValue.all_from_branch_attribute(
        BranchAttributeSpec(name='dm', long_name='deformation model', value_options=['Geodetic', 'Geologic'])
    )

    crustal_branches = FaultSystemLogicTree('Cru', 'Crustal')
    for (a, b, c, d) in itertools.product(C, s, bN, dm):
        crustal_branches.branches.append(Branch(values=[a, b, c, d], weight=0.125, onfault_nrml_id='ABC'))
    return crustal_branches


def test_fslt_example():
    fslt = build_crustal_branches()

    print(fslt)
    assert fslt.branches[-1].values[0].name == 'C'
    assert fslt.branches[-1].values[0].long_name == 'area-magnitude scaling'
    assert fslt.branches[-1].values[0].value == 4

    assert fslt.branches[-1].values[1].value == 1.0
    assert fslt.branches[-1].values[2].value == (0.95, 19.2)
    assert fslt.branches[-1].values[3].value == 'Geologic'

    assert fslt.validate_weights()


def test_full_slt():

    slt = SourceLogicTree('NSHM_1.0.0', 'initial', fault_system_lts=[build_crustal_branches()])
    print(slt)
    assert slt.fault_system_lts[0].branches[-1].values[0].name == 'C'


def test_serialise_slt():

    slt = SourceLogicTree('NSHM_1.0.0', 'initial', fault_system_lts=[build_crustal_branches()])

    slt_dict = dataclasses.asdict(slt)
    print(slt_dict)
    assert (
        slt.fault_system_lts[0].branches[-1].values[0].name
        == slt_dict['fault_system_lts'][0]['branches'][-1]['values'][0]['name']
    )


class TestSourceLogicTreeSpecification:
    def test_slt_v8(self):
        config = (
            Path(__file__).parent.parent / 'nzshm_model' / 'logic_tree' / 'source_logic_tree' / 'SLT_v8_gmm_v2_final.py'
        )
        slt = from_config(config)

        slt_spec = slt.derive_spec()

        print(slt_spec)
        assert slt_spec.fault_system_lts[0].branches[0].name == 'dm'
        assert slt_spec.fault_system_lts[0].branches[0].long_name == 'deformation model'
        assert slt_spec.fault_system_lts[0].branches[0].value_options == ['0.7']

        assert slt_spec.fault_system_lts[2].branches[0].name == 'dm'
        assert slt_spec.fault_system_lts[2].branches[0].long_name == 'deformation model'
        assert slt_spec.fault_system_lts[2].branches[0].value_options == ['geodetic', 'geologic']

    def test_large_SLT_example_A_crustal(self):
        config = Path(__file__).parent / 'fixtures' / 'large_SLT_example_A.py'
        slt = from_config(config)

        # print(slt)
        slt_spec = slt.derive_spec()

        print(slt_spec)
        assert slt_spec.fault_system_lts[0].short_name == "PUY"
        assert slt_spec.fault_system_lts[0].branches[0].name == 'dm'
        assert slt_spec.fault_system_lts[0].branches[0].long_name == 'deformation model'
        assert slt_spec.fault_system_lts[0].branches[0].value_options == ['']

        assert slt_spec.fault_system_lts[2].short_name == "CRU"
        assert slt_spec.fault_system_lts[2].branches[0].name == 'dm'
        assert slt_spec.fault_system_lts[2].branches[0].long_name == 'deformation model'
        assert slt_spec.fault_system_lts[2].branches[0].value_options == ['geologic']


class TestFlattenedSourceLogicTree:
    def setup(self):
        ba0 = BranchAttributeSpec(name='C', long_name='area-magnitude scaling', value_options=[4.0, 4.1, 4.2])
        ba1 = BranchAttributeSpec(name='C', long_name='area-magnitude scaling', value_options=[4.1, 4.2, 4.3])

        self.bavs0 = list(BranchAttributeValue.all_from_branch_attribute(ba0))
        self.bavs1 = list(BranchAttributeValue.all_from_branch_attribute(ba1))
        branches0 = [Branch([bav], 1.0 / 3.0) for bav in self.bavs0]
        branches1 = [Branch([bav], 1.0 / 3.0) for bav in self.bavs1]

        self.fault_system0 = FaultSystemLogicTree('A', 'fault system A', branches0)
        self.fault_system1 = FaultSystemLogicTree('B', 'fault system B', branches1)

    def test_flattened_slt(self):

        slt = SourceLogicTree('v1', 'slt', [self.fault_system0, self.fault_system1])
        flattened_lt = FlattenedSourceLogicTree.from_source_logic_tree(slt)
        assert len(flattened_lt.branches) == 9
        assert sum([b.weight for b in flattened_lt.branches]) == pytest.approx(1.0)

    def test_correlated_flattened_slt(self):

        correlations = [
            SourceLogicTreeCorrelation('A', 'B', [bav0], [bav1]) for bav0, bav1 in zip(self.bavs0, self.bavs1)
        ]

        slt = SourceLogicTree('v1', 'correlated slt', [self.fault_system0, self.fault_system1])
        slt.correlations = correlations
        flattened_lt = FlattenedSourceLogicTree.from_source_logic_tree(slt)

        assert len(flattened_lt.branches) == 3
        assert sum([b.weight for b in flattened_lt.branches]) == pytest.approx(1.0)
